(function(global, document, $, Milk, JSON, widgetApi, tvKeyValue, pluginApi) {

    /** Base class for scenes */
    var Scene = function() {
        this.translucent = false;
        this.panels = [];
        this._activePanel = null;
    };

    Scene.Panel = function(dom, onKeyDown, controlTips) {
        this.dom = $(dom);
        this.onKeyDown = onKeyDown;
        this.controlTips = controlTips;
        this.scene = null;
    };

    /** Class for repositories */
    var Repo = function(name) {
        this._listeners = [];
        this._isLoading = false;
        this._data = null;

        this.storagePolicy = Repo.STORE_IF_NATIVE;
        this.name = name;
    };
    Repo.STORE_ALWAYS = "store data locally in any case";
    Repo.STORE_NEVER = "do not store locally";
    Repo.STORE_IF_NATIVE = "store locally only if localStorage is supported";

    var storagePrefix = "repository-data-";



    /**
     * Loads data from the Internet.
     * @param successCallback(data) - callback to pass loaded data to
     * @param failCallback() - callback to call on fail
     */
    Repo.prototype.loadData = function(successCallback/*=function(data)*/,failCallback/*=function()*/){
        throw new Error("loadData(successCallback(data), failCallback()) must be overridden");
    };


    Scene.Panel.prototype.onActivate = function(){};
    Scene.Panel.prototype.onDeactivate = function(){};
    Scene.Panel.prototype.isActive = function() { return this.scene.getActivePanel() === this };

    Scene.prototype.isActive = function() { return App.getActiveScene() === this };

    Scene.prototype.init = function(dom,args){};
    Scene.prototype.onStart = function(resumed,finishedSceneName,finishedSceneRes){};
    Scene.prototype.onPause = function(){};
    Scene.prototype.finish = function(){}; // return scene result

    Scene.prototype.setPanels = function(panels, active) {
        var self = this;
        self.panels = (panels || []).slice();
        self.panels.forEach(function(pnl) {
            pnl.scene = self;
        });
        self.setActivePanel(typeof active === 'undefined' ? panels[0] : active);
    };
    /** Handle keyDown event. Return true if event was handled */
    Scene.prototype.onKeyDown = function(keyCode) {
        var pnl = this.getActivePanel();
        if (pnl && pnl.onKeyDown) {
            return pnl.onKeyDown(keyCode);
        } else {
            return false;
        }
    };
    Scene.prototype.getActivePanel = function() { return this._activePanel }
    Scene.prototype.setActivePanel = function(pnl) {
        if (this._activePanel) {
            this._activePanel.onDeactivate();
        }

        this.panels.forEach(function(p) {
            $(p.dom).removeClass('panel-active');
        });

        if (pnl) {
            pnl.dom.addClass('panel-active');
            this._activePanel = pnl;
            if (typeof pnl.controlTips === 'function') {
                this.setControlTipsUi(pnl.controlTips());
            } else if (pnl.controlTips) {
                this.setControlTipsUi(pnl.controlTips);
            }
            pnl.onActivate();
        }
    };

    /**
     * Utility function for rendering help panel
     * @param keymap - array of button objects like { icon:"back", text:"[0~9]", title:"" }
     * @param div - panel div to add buttons to
     */
    Scene.prototype.setControlTipsUi = function(keymap, div) {
        var self = this;
        // TODO: remove .scene-footer
        div = div ? $(div) : 
              $(self.uiControlTips).length ? self.uiControlTips : 
              $('.scene-footer', self.dom);
        // TODO: remove fields corresponding to current tips
        div.empty();
        keymap = (keymap || []).slice().reverse(); // reverse because float:right is used
        keymap.forEach(function(key) {
            var span = $('<SPAN>')
                .addClass('control-tip')
                .prepend($('<SPAN>').addClass('title').text(key.title))
                .appendTo(div);

            if (key.icon) {
                span.prepend($('<IMG>').attr('src', 'icons/'+key.icon+'.png'));
                span.attr('data-icon', key.icon);
            } else if (key.text) {
                span.prepend($('<EM>').text(key.text));
                span.attr('data-text', key.text);
            }

            if (key.id) {
                span.attr('data-id', key.id);
            }
            if (key.field) {
                key.field.split(' ').forEach(function(f) {
                    self[f] = span;
                })
            }
        });
    };

    Scene.prototype.loadingMessage = "Идёт загрузка данных...";
    Scene.prototype.noDataMessage = "Не удалось загрузить данные";

    Scene.prototype.hideMessage = function() {
		if (this._hideToastMessageTimeout) {
			clearTimeout(this._hideToastMessageTimeout);
		}
        if (this._uiMessage) {
            this._uiMessage.remove();
            this._uiMessage = null;
        }
    }

    Scene.prototype.showMessage = function(html) {
        if (!this._uiMessage) {
            this._uiMessage = $('<DIV>').addClass('overlay-info').prependTo(this.dom);
        }
        this._uiMessage.html('<span class="overlay-info-content">' + html + '</span>');
    }
	
	Scene.prototype.showToast = function(html, duration) {
		var self = this;
		self.showMessage(html);
		self._hideToastMessageTimeout = setTimeout(function() {
			self.hideMessage();
		}, duration || 1000);
	}

    /**
     * Displays one of the following:
     *   isLoading hasData
     * -   false    true   - nothing
     * -   true    false   - fullscreen loading animation
     * -   true     true   - small loading animation above screen
     * -   false   false   - full screen "no data" message
     */
    Scene.prototype.setDataState = function(isLoading, hasData) {
        if (!isLoading && hasData) {
            this.hideMessage();
        } else if (isLoading && !hasData) {
            this.showMessage(this.loadingMessage);
        } else if (isLoading && hasData) {
            // TODO: show small loading animation
        } else if (!isLoading && !hasData) {
            this.showMessage(this.noDataMessage);
        }
    }

    var App = global.App = {
        tvKeyValue: tvKeyValue,
        widgetApi: widgetApi,
		pluginApi: pluginApi,

        Scene: Scene,
        sceneClasses: {},
        registerSceneClass: function(SceneClass, name) {
            if (this.sceneClasses[name])
                throw new Error("Scene class '" + name + "' is already registered");
            this.sceneClasses[name] = SceneClass;
            SceneClass.prototype.name = name;
        },

        getActiveScene: function() {
            return this.scenes.length == 0 ? null : this.scenes[this.scenes.length-1];
        },

        Repo: Repo,
        repos: {},
        registerRepo: function(repo) {
            if (this.repos[repo.name])
                throw new Error("Repo '" + repo.name + "' is already registered");
            this.repos[repo.name] = repo;
        },

        scenes: [],
    };

    App.popScene = function() {
        var scene = this.scenes.pop();
        scene.onPause();
        var sceneRes = scene.finish();
        scene.dom.remove();
        if (this.scenes.length) {
            this.scenes[this.scenes.length-1].dom.removeClass('scene-invisible');
            this.scenes[this.scenes.length-1].onStart(true, scene.name, sceneRes);
        }
    };

    App.clearScenes = function() {
        while (this.scenes.length) {
            this.popScene();
        }
    };

    $.fn.fields = function() {
        return this.data('fields') || {};
    }

    function setField(obj, path, val) {
        var parts = path.split('.')
          , name = parts.pop();
        parts.forEach(function(part) {
            obj = obj[part] = obj[part] || {};
        })
        obj[name] = val;
    }

    function extractFields(node, target, /* internal: */ isInsideTemplate, rootNode, rootNodeFields) {
        if (!node) {
            return;
        }
        if (!rootNode || node.getAttribute('data-is-root') != null) {
            rootNode = node;
            rootNodeFields = {};
            $(rootNode).data('fields', rootNodeFields);
        }

        var templateName = node.getAttribute('data-template')
          , field = node.getAttribute('data-field')
            // build an array to not iterate through a collection
          , children = Array.prototype.filter.call(node.childNodes || node.children, function(c) {
                return c.nodeType == 1
            });

        if (templateName && field) {
            throw new Error("Templates cannot have attribute 'field'");
        }

        if (templateName && node == rootNode) {
            throw new Error("Root node cannot be a template");
        }

        if (field && !(templateName || isInsideTemplate)) {
            node.removeAttribute('data-field');
            if (target) {
                field.split(' ').forEach(function(f) {
                    setField(target, f, $(node));
                })
            }
            field.split(' ').forEach(function(f) {
                setField(rootNodeFields, f, $(node));
            });
        } // else: leave field to be handled by App.renderTemplate

        children.forEach(function(child) {
            extractFields(child, target, isInsideTemplate || templateName, rootNode, rootNodeFields);
        });

        if (templateName) {
            node.removeAttribute('data-template');
            var tmpNode = document.createElement('div');
            tmpNode.appendChild(node);
            setField(target, templateName, tmpNode.innerHTML);
        }
    }

    App.renderTemplate = function(template, data) {
        // cannot use $(Milk.render(...)): fails on 2010
        // so, create temp node and get its children
        var res = $('<DIV>').html(Milk.render(template, data)).children();
        if (res.length != 1) {
            throw new Error("Template must contain the only one node");
        }
        extractFields(res[0]);
        // cannot remove res from temp node: it loses its data fields
        return res;
    }

    App.startScene = function(name/*, arguments */) {
        var scene = new this.sceneClasses[name]()
          , args = Array.prototype.slice.call(arguments, 1)
          , dom = $('<DIV>');

		if (this.scenes.length > 0) {
			try {
				this.scenes[this.scenes.length - 1].onPause();
			} catch(e) { alert(e) }
		}

        dom.addClass('scene scene-'+name).css({'z-index': this.scenes.length * 10});
        scene.dom = dom;

        // Load scene dom model
        var fileSystem = new FileSystem()
          , f = fileSystem.openFile("scenes/" + name + ".html", "r");
        if (f != null) {
            try {
                dom.html(f.readAll());
            } catch(e) { alert(e) }
        }
        fileSystem.closeFile(f);

        extractFields(scene.dom[0], scene);
        dom.appendTo(document.body);

        try {
            args.unshift(scene.dom);
            scene.init.apply(scene, args);
        } catch (err) { alert(err) }

        if (!scene.translucent) {
            this.scenes.forEach(function(s) {
                s.dom.addClass('scene-invisible');
            });
        }
        this.scenes.push(scene);

        try {
            args.shift(); // remove dom from arguments
            args.unshift(false); // add isResumed:false
            scene.onStart.apply(scene, args);
        } catch (err) { alert(err) }
    };

    /**
     * Registers new listener for repository events
     * Listener may have some of the following methods:
     * - onLoadingStarted(repo)
     * - onLoadingSucceeded(data, repo)
     * - onLoadingFailed(repo)
     */
    Repo.prototype.registerListener = function(obj) {
        this._listeners.push(obj);
    };

    Repo.prototype.unregisterListener = function(obj) {
        this._listeners = this._listeners.filter(function(o) { o !== obj });
    };

    var notifyListeners = function(listeners, functionName/*, args*/) {
        var i, args = Array.prototype.slice.call(arguments, 2);
        for (i = 0; i < listeners.length; i++) {
            try {
                if (listeners[i][functionName])
                    listeners[i][functionName].apply(listeners[i], args);
            } catch (err) {
               alert(err);
            }
        }
    };

    Repo.prototype.isLoading = function() { return this._isLoading }

    Repo.prototype.beginLoad = function() {
        var self = this;
        if (self._isLoading)
            return;

        self._isLoading = true;

        notifyListeners(self._listeners, 'onLoadingStarted', self);

        self.loadData(
            function(data) { // onSuccess
                self.setData(data);
                self._isLoading = false;
                notifyListeners(self._listeners, 'onLoadingSucceeded', self, data);
            },
            function() { // onFail
                self._isLoading = false;
                notifyListeners(self._listeners, 'onLoadingFailed', self);
            }
        );
    };

    Repo.prototype.getData = function() {
        if (!this._data) {
            var storageString = localStorage.getItem(storagePrefix + this.name);
            if (storageString) {
                this._data = JSON.parse(storageString);
            }
        }
        return this._data;
    };

    Repo.prototype.hasData = function() {
        return this.getData() != null;
    };

    Repo.prototype.setData = function(newData) {
        this._data = newData;
        var store;
        switch(this.storagePolicy) {
            case Repo.STORE_ALWAYS:
                store = true;
                break;
            case Repo.STORE_NEVER:
                store = false;
                break;
            case Repo.STORE_IF_NATIVE:
                store = !localStorage.isNotNative;
                break;
        }
        if (store) {
            localStorage.setItem(storagePrefix + this.name, JSON.stringify(newData));
        }
    };


    // Start application

    //$(function() { // had to comment this line because script is loading dynamically and it didn't work on tv2010

    document.body.addEventListener('keydown', function(ev) {

        var scene = App.getActiveScene()
          , handled = scene && scene.onKeyDown(ev.keyCode);

        if (!handled) {
            switch(ev.keyCode) {

                case tvKeyValue.KEY_BACK:
                case tvKeyValue.KEY_RETURN:
                case tvKeyValue.KEY_PANEL_RETURN:
                    if (App.scenes.length > 0) {
                        App.popScene();
                    }
                    // If there is no more scenes after popping - exit
                    if (App.scenes.length == 0) {
                        widgetApi.sendReturnEvent();
                    /*} else {
                        widgetApi.blockNavigation(ev);*/
                    }
                    handled = true;
                    break;

                case tvKeyValue.KEY_EXIT:
                    App.clearScenes();
                    widgetApi.sendExitEvent();

                    handled = true;
                    break;

            }
        }

        if (handled) {
            ev.preventDefault();
            ev.stopPropagation();
            widgetApi.blockNavigation(ev);
            return false;
        //} else {
        //    return true;
        }
    }, true);

        //widgetApi.sendReadyEvent();

    //});

    App.define = {
		PL_AUDIO_OUTPUT_DEVICE_EXTERNAL:3,
		PL_AUDIO_OUTPUT_DEVICE_RECEIVER:4,
		PL_AUDIO_VOLUME_KEY_UP:0,
		PL_AUDIO_VOLUME_KEY_DOWN:1,

		PLR_FALSE:0,
		PLR_TRUE:1,

		PL_AUDIO_MUTE_OFF:1,
		PL_AUDIO_MUTE_ON:0,

		PL_TV_PRODUCT_TYPE_TV:0,
		PL_TV_PRODUCT_TYPE_MONITOR:1,
		PL_TV_PRODUCT_TYPE_BD:2,

		PL_CMN_INFO_VERSION:0,

		PL_NNAVI_STATE_BANNER_NONE:0,
		PL_NNAVI_STATE_BANNER_VOL:1,
		PL_NNAVI_STATE_BANNER_VOL_CH:2,

		EVENT_TO_DEVICE_SCREEN_SAVER_ON:3,
		EVENT_TO_DEVICE_SCREEN_SAVER_OFF:4,
		EVENT_RETURN_WIDGETID:"100",
		EVENT_RETURN_TO_WIDGET:"101",

		PL_PRFID_AUTO_PROTECTION_TIME:13,

		PICTURE_SETTING:1,
		SOUND_SETTING:0,
		PL_DTVAPP_TOOL:18,
		EVENT_TO_DEVICE_HIDE_TOOLS:6,

		GET_SNS_LIST:"32",
		GET_APP_ALL_LIST:"36",
		GET_APP_PATH:"33",

		MOVE_APP_INFO:"34",
		PL_TVMW_SOURCE_MEDIA:43
	};
    
    App.debug = function(message) {
        if (!App._debugUi) {
            App._debugUi = $('<DIV>').css({
                background: 'White',
                color: 'Black',
                zIndex: 2000,
                position: 'absolute',
                top: '0',
                left: '0'
            }).appendTo(document.body);
            App._debugCount = 0;
        }
        if (App._debugCount == 0 || App._debugCount > 10) {
            App._debugUi.html(message);
            App._debugCount = 1;
        } else {
            App._debugUi.append("<br/>" + message);
            App._debugCount++;
        }
    }

})(this, document, jQuery, Milk, JSON, new Common.API.Widget(), new Common.API.TVKeyValue(), new Common.API.Plugin());
